function [Ainv,logLikelihood] = lca(X, max_iter, reg, tol)

% [Ainv,logLikelihood] = lca(X, max_iter, reg, tol) finds the inverse metric Ainv so as to
% maximize the leave-one-out log-probability of the data using Parzen Windows
% with that metric.
%
%% INPUTS %%
% - X contains one datapoint per column.
% - max_iter is the maximum number of updates.
% - reg is the regularization parameter.
% - tol is the minimum improvement in log-likelihood under which convergence is deemed attained.
%
%% OUTPUTS %%
% - Ainv is the local covariance matrix (takes the inverse of its Cholesky decomposition to tranform your data).
% - logLikelihood is the log-likelihood of the parameters for each datapoint.

[n_dims, n_data] = size(X);

if nargin < 2, max_iter = 100; end
if nargin < 3, reg = 10^-4; end
if nargin < 4, , tol = 10^-10; end

% Initialize the covariance matrix to the global covariance of the data.
Ainv = cov(X') + reg*eye(n_dims); % Ainv = A^{-1}.

meanlogLikelihood = -inf;

% The last iteration is here to compute the final log-likelihood.
for iter = 1:(max_iter+1)

	% The matrix logK is defined as:
	% logK_ij = -.5 (x_i - x_j)^T Ainv^(-1) (x_i - x_j)
	% logK_ii = -inf
	XAX = X'*(Ainv\X);
	dXAX = repmat(diag(XAX), 1, n_data);
	
	logK = -.5*(dXAX + dXAX' -2*XAX);

	% Each datapoint cannot use itself.
	logK(1:(n_data+1):end) = -inf;

	% Removing the top component of each row to avoid overflows.
	maxLogK = max(logK, [], 2);
	logK = logK - repmat(maxLogK, 1, cols(logK));

	% K_ij = \lambda_ij in the paper.
	K = exp(logK);
	sumK = sum(K, 2);
	logLikelihood = -.5*logdet(Ainv) + log(sumK) + maxLogK - log(n_data-1) - n_dims/2*log(2*pi);

	oldMeanlogLikelihood = meanlogLikelihood;
	meanlogLikelihood = mean(logLikelihood);
	fprintf('Iteration %i		log-likelihood = %g\n', iter-1, meanlogLikelihood);

	% At the last iteration, we do not update Ainv.
	if iter < max_iter + 1

		% Determine if enough progress has been made.
		if (meanlogLikelihood - oldMeanlogLikelihood) < tol
			fprintf('Convergence reached!\n'); break, end;

		% Row-normalize the responsibilities.
		B = K./repmat(sumK, 1, n_data);

		sumB = sum(B, 1)' + sum(B, 2);
		Ainv = (X*(diag(sumB) - B - B')*X')/n_data + reg*eye(n_dims);

	end

end
