function [B_G, B_L, logLikelihood] = lca_gauss(X, max_iter, reg, tol, B_G, B_L)

%% INPUTS %%
% - X contains one datapoint per column.
% - max_iter is the number of updates to the number of dimensions modeled by Parzen.
% - reg is the regularization on the covariance (not precision) matrices.
% - tol is the minimum improvement in log-likelihood under which convergence is deemed attained.
% - If provided, B_G and B_L are initializations for the two matrices.
%
%% OUTPUTS %%
% - B_L is square root of the covariance matrix used for Parzen windows.
% - B_G is square root of the covariance matrix used for the Gaussian.
% - logLikelihood is the log-likelihood of the parameters for each datapoint.

n_dims = size(X, 1);
n_data = size(X, 2);

% This is the optimal mean for the Gaussian and does not need to be updated.
mu = mean(X, 2);

if nargin < 2, max_iter = 100; end
if nargin < 3, reg = 10^-4; end
if nargin < 4, tol = 10^-10; end

% Initialize the variables.

if numel(reg) == 2
	reg_cl = reg(1);
	reg_cg = reg(2);
else
	reg_cl = reg;
	reg_cg = reg;
end

% C_G is the global covariance of the data.
C_G = cov(X',1) + reg_cg*eye(n_dims);
[U_G, D_G] = eig(C_G);

B_init = U_G*diag(diag(D_G).^(-.5));

% invsq_C_G is equal to C_G^{-1/2} .
invsq_C_G = B_init*U_G';

if ~exist('B_L', 'var')
	B_L = B_init;
	B_G = zeros(n_dims, 0);
end

above_one = zeros(0);
below_one = 1:n_dims;

meanlogLikelihood = -inf;

% The last iteration is here to compute the final log-likelihood.
for iter = 1:max_iter+1

	% At each step, we need to decide which dimensions are modeled by the
	% Gaussian and which dimensions are modeled by Parzen.

	% B_G is of size n_dims by p1.
	% B_L is of size n_dims by (n_dims - p1).
	L_X = B_L'*X; % Of size (n_dims - p1) * n_data
	G_X = B_G'*(X - repmat(mu, 1, n_data)); % Of size p1 * n_data

	XAX = L_X'*L_X;
	dXAX = repmat(diag(XAX), 1, n_data);
	
	% We do not need to include the Gaussian model as it will vanish in the row normalization.
	logK = -.5*(dXAX + dXAX' -2*XAX);

	% Each datapoint cannot use itself.
	logK(1:(n_data+1):end) = -inf;

	% Remove the top element for each row. This is to avoid overflows.
	maxLogK = max(logK, [], 2);
 logK = logK - repmat(maxLogK, 1, cols(logK));

	K = exp(logK);
	sumK = sum(K, 2);

	logLikelihood = .5*logdet(B_L*B_L' + B_G*B_G') + log(sumK) + maxLogK -.5*sum(G_X.^2, 1)' - log(n_data-1) - n_dims/2*log(2*pi);
	oldMeanlogLikelihood = meanlogLikelihood;
	meanlogLikelihood = mean(logLikelihood);
	fprintf('Iteration %i		log-likelihood = %g n_local = %g\n', iter-1, meanlogLikelihood, cols(B_L));

	% At the last iteration, we do not update Ainv.
	if iter < max_iter + 1

		% Determine if enough progress has been made.
		if (meanlogLikelihood - oldMeanlogLikelihood) < tol
			fprintf('Convergence reached!\n'); break, end;

		% Compute the optimal EM coefficients.
		L = K./repmat(sumK, 1, n_data);

		% Recompute the local covariance matrix.
		sumB = sum(L, 1)' + sum(L, 2);
		C_L = (X*(diag(sumB) - L - L')*X')/n_data + reg_cl*eye(n_dims);

		% Compute the eigenvectors and eigenvalues of C_G^{-1/2} C_L C_G^{-1/2}.
		C_GL = invsq_C_G*C_L*invsq_C_G;

		% Symmetrize the matrix to avoid numerical inaccuracies.
		C_GL = .5*(C_GL + C_GL');

		[V,D] = eig(C_GL);
		list_eigen = diag(D);

		% Find which dimensions must be modelled by Parzen windows. These are the dimensions with eigenvalues smaller than one.
		above_one = find(list_eigen > 1);
		below_one = find(list_eigen <= 1);
		B_L = invsq_C_G*V(:, below_one)*diag(list_eigen(below_one).^(-.5));
		B_G = invsq_C_G*V(:, above_one);

	end

end
