function [best_B_G, best_B_L, best_logProb] = lca_gauss_red(X, iter_lca_gauss, reg, nRecompute)

% lca_gauss_red tries to find the number of dimensions modeled by Parzen which
% yields the highest training log-likelihood. Since the standard approach
% (coded in lca_gauss) may sometimes get stuck in local minima, we force
% from time to time the number of dimensions modeled by Parzen to change using
% a dichotomic search.
%
% - X is the dataset containing one datapoint per column.
%
% - iter_lca_gauss is the maximum number of assignment updates (that is,
% dimensions which go from global to local or vice-versa) in lca_gauss
%
% - reg is the regularization parameter in lca_gauss
%
% - nRecompute is the number of iterations during which B_G and B_L are
% optimized between two assignment updates
%
% Start with all dimensions as local.

n_dims = size(X, 1);

% If iter_lca_gauss is a vector of size 2, the first element is the number of
% iterations for the intermediate steps and the second one is the number of
% iterations for the final step.

if numel(iter_lca_gauss) == 2
	iter_final = iter_lca_gauss(2);
	iter_lca_gauss = iter_lca_gauss(1);
end

% Compute an upper bound on the number of dimensions modelled by Parzen.
[B_G, B_L, logProb] = lca_gauss(X, iter_lca_gauss, reg, n_dims, nRecompute, 1);

right_local = cols(B_L);
right_logProb = mean(logProb);

right_B_G = B_G;
right_B_L = B_L;

% Compute a lower bound on the number of dimensions modelled by Parzen.
% The left logProb is the one with only Gaussian components.
[B_G, B_L, logProb] = lca_gauss(X, 1, reg, 0, 1, 1);

left_local = cols(B_L);
left_logProb = mean(logProb);
left_B_G = B_G;
left_B_L = B_L;

if left_local == right_local
  fprintf('Optimal number of components found. Performing final optimization.\n');
		[B_G, B_L, logProb] = lca_gauss(X, iter_final, reg, {right_B_G, right_B_L}, nRecompute, 1);
  best_B_G = B_G;
  best_B_L = B_L;
  best_logProb = mean(logProb);
		return;
end

% Give some dimensions from local to global.
if right_local - left_local == 2
	new_local = left_local + 1;
else
	new_local = floor(left_local + (right_local - left_local)/2);
end
new_B_L = right_B_L(:, 1:new_local);
new_B_G = [right_B_L(:,new_local+1:end) right_B_G];

% Compute the third point.
[B_G, B_L, logProb] = lca_gauss(X, iter_lca_gauss, reg, {new_B_G, new_B_L}, nRecompute, 0);

mid_logProb = mean(logProb);
mid_local = cols(B_L);
mid_B_G = B_G;
mid_B_L = B_L;

% We may now start to iterate.

% To avoid infinite loops, the algorithm stops as soon as the optimization
% procedure ends on a point visited before.
list_visited = [left_local mid_local right_local];

while 1

	if ((right_logProb > mid_logProb) && (mid_logProb > left_logProb))

		% The right one is the best.
		best_logProb = right_logProb;
		best_local = right_local;
		best_B_G = right_B_G;
		best_B_L = right_B_L;

		% If all the states between mid_local and right_local have already been
		% visited, the optimization is over.
		if ismember(mid_local:right_local, list_visited)
			fprintf('Optimal number of components found. Performing final optimization.\n');
			[B_G, B_L, logProb] = lca_gauss(X, iter_final, reg, {best_B_G, best_B_L}, nRecompute, 1);
			best_logProb = mean(logProb);
			best_B_G = B_G;
			best_B_L = B_L;
			return;
			end

		left_local = mid_local;
	 left_logProb = mid_logProb;
  left_B_G = mid_B_G;
  left_B_L = mid_B_L;

		% The number of dimensions to try is halfway between the two bounds.
		not_tried = setdiff(mid_local:right_local, list_visited);
		new_local = not_tried(ceil(numel(not_tried)/2));

		new_B_L = right_B_L(:, 1:new_local);
		new_B_G = [right_B_L(:,new_local+1:end) right_B_G];
		[B_G, B_L, logProb] = lca_gauss(X, iter_lca_gauss, reg, {new_B_G, new_B_L}, nRecompute, 0);
		list_visited = [list_visited new_local];
		mid_logProb = mean(logProb);
		mid_local = cols(B_L);
		mid_B_G = B_G;
		mid_B_L = B_L;

	elseif ((right_logProb <= mid_logProb) && (mid_logProb <= left_logProb))

  % The left one is the best.
		best_logProb = left_logProb;
		best_local = left_local;
		best_B_G = left_B_G;
		best_B_L = left_B_L;

		if ismember(left_local:mid_local, list_visited)
			fprintf('Optimal number of components found. Performing final optimization.\n');
			[B_G, B_L, logProb] = lca_gauss(X, iter_final, reg, {best_B_G, best_B_L}, nRecompute, 1);
			best_logProb = mean(logProb);
			best_B_G = B_G;
			best_B_L = B_L;
			return;
		end

		right_local = mid_local;
		right_logProb = mid_logProb;
		right_B_G = mid_B_G;
		right_B_L = mid_B_L;

		not_tried = setdiff(left_local:mid_local, list_visited);
		new_local = not_tried(ceil(numel(not_tried)/2));

		new_B_L = [best_B_L best_B_G(:, 1:(new_local-left_local))];
		new_B_G = best_B_G(:,(new_local - left_local + 1):end);

		[B_G, B_L, logProb] = lca_gauss(X, iter_lca_gauss, reg, {new_B_G, new_B_L}, nRecompute, 0);
		list_visited = [list_visited new_local];
		mid_logProb = mean(logProb);
		mid_local = cols(B_L);
		mid_B_G = B_G;
		mid_B_L = B_L;

	else

		% The middle one is the best.
		best_logProb = mid_logProb;
		best_local = mid_local;
		best_B_G = mid_B_G;
		best_B_L = mid_B_L;

		if ismember(left_local:right_local, list_visited)
			fprintf('Optimal number of components found. Performing final optimization.\n');
			[B_G, B_L, logProb] = lca_gauss(X, iter_final, reg, {best_B_G, best_B_L}, nRecompute, 1);
			best_logProb = mean(logProb);
			best_B_G = B_G;
			best_B_L = B_L;
			return;
		end

		% Create another mid one and compare the two.
		not_tried = setdiff(left_local:right_local, list_visited);
		new_local = not_tried(ceil(numel(not_tried)/2));
	
		if new_local > mid_local
			new_B_L = [mid_B_L mid_B_G(:, 1:(new_local-mid_local))];
			new_B_G = mid_B_G(:,(new_local - mid_local + 1):end);
		else
			new_B_L = mid_B_L(:, 1:new_local);
			new_B_G = [mid_B_L(:,new_local+1:end) mid_B_G];
		end

  [B_G, B_L, logProb] = lca_gauss(X, iter_lca_gauss, reg, {new_B_G, new_B_L}, nRecompute, 0);
		list_visited = [list_visited new_local];
		new_logProb = mean(logProb);
		new_B_G = B_G;
		new_B_L = B_L;
		
		if new_logProb > mid_logProb

   best_logProb = new_logProb;
   best_local = new_local;
   best_B_G = new_B_G;
   best_B_L = new_B_L;

			if new_local > mid_local

	   left_local = mid_local;
	   left_logProb = mid_logProb;
	   left_B_G = mid_B_G;
	   left_B_L = mid_B_L;

			else

    right_local = mid_local;
    right_logProb = mid_logProb;
    right_B_G = mid_B_G;
    right_B_L = mid_B_L;

			end

   mid_local = new_local;
   mid_logProb = new_logProb;
   mid_B_G = new_B_G;
   mid_B_L = new_B_L;

		else
   % The old one is the best.
   best_logProb = mid_logProb;
   best_local = mid_local;
   best_B_G = mid_B_G;
   best_B_L = mid_B_L;

			if new_local > mid_local

				right_local = new_local;
	   right_logProb = new_logProb;
	   right_B_G = new_B_G;
	   right_B_L = new_B_L;

			else

    left_local = new_local;
    left_logProb = new_logProb;
    left_B_G = new_B_G;
    left_B_L = new_B_L;
				
			end

			new_local = mid_local;

		end

	end

end
