function [Ainv,logProb] = lca_online(X, max_iter, reg, gamma_iter, batch_size, neighborhood_size, Ainv)

% Ainv = lca_online(X, max_iter, reg, gamma_iter, batch_size, neighborhood_size, Ainv) finds the
% inverse metric Ainv so as to maximize the leave-one-out log-probability of
% the data using Parzen Windows with Ainv metric.
%
% The last argument is optional and allows the user to pursue the
% optimization.
%
% X contains one datapoint per column.

[n_dims, n_data] = size(X);

if nargin < 2, max_iter = 100; end
if nargin < 3, reg = 10^-4; end

n_batches = ceil(n_data/batch_size);

% We use exponential weight decay. Thus, the Ainv found at iteration t is still
% present with some weight at iteration t+1. Its weight is exactly
% gamma^n_batches. gamma_iter defines this value. Thus, to retrieve gamma out
% of gamma_iter, we must compute:
gamma = gamma_iter^(1/n_batches);

if nargin < 7

	% Initialize the variables.
	Ainv = cov(X') + reg*eye(n_dims); % Ainv = A^{-1}.

end

% If no neighborhood_size is set, use all datapoints.
if isempty(neighborhood_size) || neighborhood_size > n_data , neighborhood_size = n_data; end

logProb = zeros(n_data, 1);

for iter = 1:max_iter

	index = 0;
	
	for iBatch = 1:n_batches

		% Datapoints for which you will compute the local covariance matrix.
		if index + batch_size <= n_data
			X_batch = X(:, index + (1:batch_size));
			index_batch = index + (1:batch_size);
		else
			X_batch = X(:, index + 1:end);
			index_batch = (index + 1):n_data;
		end

		% How many points in the batch? The only issue is with the last batch.
		n_x = cols(X_batch);

		% The matrix B is defined as:
		% B_ij = exp(-.5 (x_i - x_j)^T A (x_i - x_j)) .

		% If the neighborhood is the entire dataset, we know we do not have to compute the batches.
		if neighborhood_size == n_data
			Y = X;
			AinvY = Ainv\Y;
		 AinvX = AinvY(:, index_batch);
		else
		 AinvX = Ainv\X_batch;

			% We need to pick the neighbors at random. If the batch is not the whole
			% dataset, we try to pick neighbors which are not in the patch.

			R = randperm(n_data);

			if batch_size == n_data
				index_neighbors = R(1:neighborhood_size);
				Y = X(:, index_neighbors);
				AinvY = AinvX(:, index_neighbors);
			else
				if neighborhood_size + n_x <= n_data
					R = setxor(R, index_batch);
				end
				index_neighbors = R(1:neighborhood_size);
				Y = X(:, index_neighbors);
				AinvY = Ainv\Y;
			end
		end

		if neighborhood_size == n_data

			YAY = sum(Y.*AinvY, 1);
			XAX = YAY(index_batch);

		elseif batch_size == n_data

			XAX = sum(X.*AinvX, 1);
			YAY = XAX(index_neighbors);
	
		else
	
			XAX = sum(X_batch.*AinvX, 1);
			YAY = sum(Y.*AinvY, 1);

		end

		XAY = X_batch'*AinvY;
		dXAX = repmat(XAX', 1, neighborhood_size);
		dYAY = repmat(YAY, n_x, 1); 
	
		logK = -.5*(dXAX + dYAY -2*XAY);

		% Each point cannot use itself in its neighborhood.

		if neighborhood_size == n_data
	
			% If we use all datapoints as neighbors, we know which ones should be
			% removed.  This is the elements logK(i, index + i). Since Matlab lists the
			% elements of a matrix columnwise, we start at (n_x * index + 1) and end at
			% (n_x * (index + n_x)).

			logK((n_x*index + 1):(n_x+1):(n_x * (index + n_x))) = -inf;

			% The denominator is the log of the number of neighbors for each datapoint.
			denominator = log(n_data - 1);

		elseif (neighborhood_size + n_x <= n_data)

			% In that case, we used non-intersecting batch and neighbors.

			denominator = log(neighborhood_size);

		else

			% For all the other cases, this is a bit too complicated and we shall resort
			% to a simple but inefficient way of doing it.

			is_itself = repmat(index_batch', 1, neighborhood_size) == repmat(index_neighbors, n_x, 1);
			logK(is_itself) = -inf;
			denominator = log(sum(1-is_itself, 2));

		end

		maxLogK = max(logK, [], 2);
		logK = logK - repmat(maxLogK, 1, cols(logK));
		expLogK = exp(logK);
		sumExpLogK = sum(expLogK, 2);

		logProb(index_batch) = -.5*logdet(Ainv) + log(sumExpLogK) + maxLogK - denominator - n_dims/2*log(2*pi);

		B = expLogK./repmat(sumExpLogK, 1, neighborhood_size);
	
		k1 = sum(B, 1);
		k2 = ones(n_x, 1);
	
		XBY = X_batch*B*Y';
		Xk = X_batch.*repmat(k2', n_dims, 1);
		Yk = Y.*repmat(k1, n_dims, 1);
		Ainv = gamma*Ainv + (1-gamma)*(Xk*X_batch' + Yk*Y' - XBY - XBY')/n_x + (1-gamma)*reg*eye(n_dims);

		index = index + batch_size;

	end

	meanLogProb = mean(logProb);
	fprintf('Iteration %i		log-prob = %g\n', iter, meanLogProb);

end
